/*
    +--------------------------------------------------------------------+
    | libmemcached-awesome - C/C++ Client Library for memcached          |
    +--------------------------------------------------------------------+
    | Redistribution and use in source and binary forms, with or without |
    | modification, are permitted under the terms of the BSD license.    |
    | You should have received a copy of the license in a bundled file   |
    | named LICENSE; in case you did not receive a copy you can review   |
    | the terms online at: https://opensource.org/licenses/BSD-3-Clause  |
    +--------------------------------------------------------------------+
    | Copyright (c) 2006-2014 Brian Aker   https://datadifferential.com/ |
    | Copyright (c) 2020-2021 Michael Wallner        https://awesome.co/ |
    +--------------------------------------------------------------------+
*/

#pragma once

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {
  vbucket_state_active = 1, /**< Actively servicing a vbucket. */
  vbucket_state_replica,    /**< Servicing a vbucket as a replica only. */
  vbucket_state_pending,    /**< Pending active. */
  vbucket_state_dead        /**< Not in use, pending deletion. */
} vbucket_state_t;

#define is_valid_vbucket_state_t(state) \
  (state == vbucket_state_active || state == vbucket_state_replica \
   || state == vbucket_state_pending || state == vbucket_state_dead)

#ifdef __cplusplus
}
#endif
